# This dummy django settings file is used by sphinx while loading
# eulcore.django.* to examine it for autodoc generation.
FEDORA_ROOT = 'dummy root'  # NOTE: there are documentation errors without this
